@@ -1189,7 +1189,6 @@ g_local_file_set_display_name (GFile         *file,
       if (errsv != ENOENT)
         {
           g_set_io_error (error, _("Error renaming file %s: %s"), new_file, errsv);
-          g_object_unref (new_file);
           return NULL;
         }
     }
@@ -1197,7 +1196,6 @@ g_local_file_set_display_name (GFile         *file,
     {
       g_set_error_literal (error, G_IO_ERROR, G_IO_ERROR_EXISTS,
                            _("Can’t rename file, filename already exists"));
-      g_object_unref (new_file);
       return NULL;
     }
 
@@ -1999,6 +1997,79 @@ _g_local_file_is_lost_found_dir (const char *path, dev_t path_dev)
 }
 #endif
 
+/* Check whether subsequently deleting the original file from the trash
+ * (in the gvfsd-trash process) will succeed. If we think it won’t, return
+ * an error, as the trash spec says trashing should not be allowed.
+ * https://specifications.freedesktop.org/trash-spec/latest/#implementation-notes
+ *
+ * Check ownership to see if we can delete. gvfsd will automatically chmod
+ * a file to allow it to be deleted, so checking the permissions bitfield isn’t
+ * relevant.
+ */
+static gboolean
+check_removing_recursively (GFile        *file,
+                            gboolean      user_owned,
+                            uid_t         uid,
+                            GCancellable *cancellable,
+                            GError       **error)
+{
+  GFileEnumerator *enumerator;
+
+  enumerator = g_file_enumerate_children (file,
+                                          G_FILE_ATTRIBUTE_STANDARD_TYPE ","
+                                          G_FILE_ATTRIBUTE_UNIX_UID,
+                                          G_FILE_QUERY_INFO_NOFOLLOW_SYMLINKS,
+                                          cancellable,
+                                          error);
+
+  if (!enumerator)
+    return FALSE;
+
+  while (TRUE)
+    {
+      GFileInfo *info;
+      GFile *child;
+
+      if (!g_file_enumerator_iterate (enumerator, &info, &child, cancellable, error))
+        {
+          g_object_unref (enumerator);
+          return FALSE;
+        }
+
+      if (!info)
+        break;
+
+      if (!user_owned)
+        {
+          GLocalFile *local = G_LOCAL_FILE (child);
+
+          g_set_error (error, G_IO_ERROR, G_IO_ERROR_PERMISSION_DENIED,
+                       _("Unable to trash child file %s"), local->filename);
+          g_object_unref (enumerator);
+          return FALSE;
+        }
+
+      if ((g_file_info_get_file_type (info) == G_FILE_TYPE_DIRECTORY))
+        {
+          uid_t fuid;
+
+          fuid = g_file_info_get_attribute_uint32 (info,
+                                                   G_FILE_ATTRIBUTE_UNIX_UID);
+          if (!check_removing_recursively (child,
+                                           fuid == uid,
+                                           uid,
+                                           cancellable,
+                                           error))
+            {
+              g_object_unref (enumerator);
+              return FALSE;
+            }
+        }
+    }
+  g_object_unref (enumerator);
+  return TRUE;
+}
+
 static gboolean
 g_local_file_trash (GFile         *file,
 		    GCancellable  *cancellable,
@@ -2006,6 +2077,7 @@ g_local_file_trash (GFile         *file,
 {
   GLocalFile *local = G_LOCAL_FILE (file);
   GStatBuf file_stat, home_stat;
+  dev_t checked_st_dev;
   const char *homedir;
   char *trashdir, *topdir, *infodir, *filesdir;
   char *basename, *trashname, *trashfile, *infoname, *infofile;
@@ -2051,6 +2123,8 @@ g_local_file_trash (GFile         *file,
   is_homedir_trash = FALSE;
   trashdir = NULL;
 
+  checked_st_dev = file_stat.st_dev;
+
   /* On overlayfs, a file's st_dev will be different to the home directory's.
    * We still want to create our trash directory under the home directory, so
    * instead we should stat the directory that the file we're deleting is in as
@@ -2058,13 +2132,14 @@ g_local_file_trash (GFile         *file,
    */
   if (!S_ISDIR (file_stat.st_mode))
     {
+      GStatBuf parent_stat;
       path = g_path_get_dirname (local->filename);
       /* If the parent is a symlink to a different device then it might have
        * st_dev equal to the home directory's, in which case we will end up
        * trying to rename across a filesystem boundary, which doesn't work. So
        * we use g_stat here instead of g_lstat, to know where the symlink
        * points to. */
-      if (g_stat (path, &file_stat))
+      if (g_stat (path, &parent_stat))
 	{
 	  errsv = errno;
 	  g_free (path);
@@ -2074,10 +2149,11 @@ g_local_file_trash (GFile         *file,
 			  file, errsv);
 	  return FALSE;
 	}
+      checked_st_dev = parent_stat.st_dev;
       g_free (path);
     }
 
-  if (file_stat.st_dev == home_stat.st_dev)
+  if (checked_st_dev == home_stat.st_dev)
     {
       is_homedir_trash = TRUE;
       errno = 0;
@@ -2379,9 +2455,22 @@ g_local_file_trash (GFile         *file,
 
   g_clear_pointer (&data, g_free);
 
-  /* TODO: Maybe we should verify that you can delete the file from the trash
-   * before moving it? OTOH, that is hard, as it needs a recursive scan
-   */
+  if (S_ISDIR (file_stat.st_mode))
+    {
+      uid_t uid = geteuid ();
+
+      if (file_stat.st_uid == uid &&
+          !check_removing_recursively (file, TRUE, uid, cancellable, error))
+        {
+          g_unlink (infofile);
+
+          g_free (filesdir);
+          g_free (trashname);
+          g_free (infofile);
+
+          return FALSE;
+        }
+    }
 
   trashfile = g_build_filename (filesdir, trashname, NULL);
 
