@@ -51,11 +51,15 @@ struct _GString
 
 GLIB_AVAILABLE_IN_ALL
 GString*     g_string_new               (const gchar     *init);
+GLIB_AVAILABLE_IN_2_78
+GString*     g_string_new_take          (gchar           *init);
 GLIB_AVAILABLE_IN_ALL
 GString*     g_string_new_len           (const gchar     *init,
                                          gssize           len);
 GLIB_AVAILABLE_IN_ALL
 GString*     g_string_sized_new         (gsize            dfl_size);
+GLIB_AVAILABLE_IN_2_86
+GString     *g_string_copy              (GString         *string);
 GLIB_AVAILABLE_IN_ALL
 gchar*      (g_string_free)             (GString         *string,
                                          gboolean         free_segment);
@@ -64,6 +68,8 @@ gchar*       g_string_free_and_steal    (GString         *string) G_GNUC_WARN_UN
 
 #if G_GNUC_CHECK_VERSION (2, 0) && (GLIB_VERSION_MIN_REQUIRED >= GLIB_VERSION_2_76)
 
+#if !defined(__cplusplus) || !G_GNUC_CHECK_VERSION (6, 1) || G_GNUC_CHECK_VERSION (7, 3)
+
 #define g_string_free(str, free_segment)        \
   (__builtin_constant_p (free_segment) ?        \
     ((free_segment) ?                           \
@@ -72,6 +78,8 @@ gchar*       g_string_free_and_steal    (GString         *string) G_GNUC_WARN_UN
     :                                           \
     (g_string_free) ((str), (free_segment)))
 
+#endif /* !defined(__cplusplus) || !G_GNUC_CHECK_VERSION (6, 1) || G_GNUC_CHECK_VERSION (7, 3) */
+
 #endif /* G_GNUC_CHECK_VERSION (2, 0) && (GLIB_VERSION_MIN_REQUIRED >= GLIB_VERSION_2_76) */
 
 GLIB_AVAILABLE_IN_2_34
